#!/usr/bin/perl -w
;########################################################################
;###	Google Calendar Proxy
;###			Programmed by Piroli YUKARINOMIYA (MagicVox)
;###			Open MagicVox.net - http://www.magicvox.net/

use vars qw( $VERSION );
$VERSION = '0.10';

use strict;
use CGI;
use LWP::UserAgent;
use XML::Simple;
use XML::LibXML;
use XML::LibXSLT;
use Digest::MD5 qw( md5_hex );
;#use CGI::Carp qw(fatalsToBrowser);#DEBUG
;#use Data::Dumper;#DEBUG



;########################################################################

;# Specify the length of URL hash keys (max=32)
use constant HASH_KEY_LENGTH =>			8;

;# Customize for your directory setting
use constant DIR_FEED =>				'./cache';
use constant DIR_XSLT =>				'./xslt';

;# Get the latest feed with HTTP connecting when a cache is expired
use constant CACHE_EXPIRE =>			60 * 30;

;# Timeout seconds using in LWP::UserAgent
use constant LWP_TIMEOUT => 			30;

;# Default mode if <mode> is not specified
use constant DEFAULT_MODE =>			'index';

;# Using when generating URL hash keys, but not very important ;)
use constant HASH_MAGICWORD =>			'THE SECRET OF BLUE WATER';



;### No need editing belows cause of internal using constant values
use constant XMLNS_GGLCAL_FUNC =>		'http://www.magicvox.net/pub/xmlns/gglcal';
use constant LWP_USER_AGENT =>			'GGLCAL/'. $VERSION;

;########################################################################
;###	Main Loop

;### Parameters
my $q = new CGI
	or errorNoContents ("Fatal Error: CGI initialize failure");

;### Check parameters - mode
my $q_mode = $q->param ('mode') || DEFAULT_MODE;
errorNoContents ("Invalid Parameter - invalid <mode> is specified")
	unless defined $q_mode && $q_mode =~ m/[\w-]./;
errorNoContents ("Invalid Parameter - There is not such a mode specified by <mode>")
	unless -e &getFilenameXSL ($q_mode);

;### Check parameters - next
my $q_next = $q->param ('next');
if (defined $q_next) {
	my $len = HASH_KEY_LENGTH;
;#
	$q_next = lc $q_next;
	errorNoContents ("Invalid Parameter - next")
		unless $q_next =~ m/^[0-9a-f]{$len}$/;
}

;### Check parameters - self
my $q_self = $q->param ('self');
if (defined $q_self) {
	my $len = HASH_KEY_LENGTH;
;#
	$q_self = lc $q_self;
	errorNoContents ("Invalid Parameter - self")
		unless $q_self =~ m/^[0-9a-f]{$len}$/;
}

$q_next = $q_self if defined $q_self && ! defined $q_next;
$q_self = $q_next if defined $q_next && ! defined $q_self;

;### Main Switch
if (defined $q_next && defined $q_self) {
	if (isExpireFeed ($q_next)) {
		my $feed_url;
		if (! defined ($feed_url = retrieveURL ($q_next, $q_self))) {
			errorNoContents ("Invalid Parameter: link not found");
		}
		my $new_feed = LWPGetFeed ($feed_url);
		saveFeed ($q_next, \$new_feed);
	}
}
else {
	my $feed_url;
	if (! defined ($feed_url = getStartupURL ())) {
		errorNoContents ("Fatal Error: not initialized");
	}
	$q_next = getURLHash ($feed_url);
	if (isExpireFeed ($q_next)) {
		my $new_feed = LWPGetFeed ($feed_url);
		saveFeed ($q_next, \$new_feed);
	}
}

;### Generate output
print "Content-type: text/html; charset=utf-8\n\n";
print &createOutput ($q_next, $q_mode);
exit;



;### errorNoContents
sub errorNoContents {
	print "Content-type: text/plain; charset=utf-8\n\n@_";
	exit;
}

;########################################################################
;###	Misc Procedures

;### Retrieve filename of XML & XSL
sub getFilenameXML { my ($hash) = @_; return $hash ? DIR_FEED. '/'. $hash. '.xml' : undef; }
sub getFilenameXSL { my ($name) = @_; return $name ? DIR_XSLT. '/'. $name. '.xsl' : undef; }

;### Check whether specified file is not existing and expired.
sub isExpireFeed {
	my ($hash) = @_;
	errorNoContents ("Program Exception: isExpireFeed [1]")
		unless defined $hash;
;#
	my $filename = getFilenameXML ($hash);
	return 1 unless -e $filename;
	return 1 unless time () < (stat ($filename))[9] + CACHE_EXPIRE;
	return 0;
}

;### Retrieve latest feed with LWP
sub LWPGetFeed {
	my ($url) = @_;
	errorNoContents ("Program Exception: LWPGetFeed [1]")
		unless defined $url;
;#
	my $ua = LWP::UserAgent->new ()
		or errorNoContents ("Fatal Error: LWP::UserAgent initialize failure");
	$ua->timeout (LWP_TIMEOUT) if LWP_TIMEOUT;
	$ua->agent (LWP_USER_AGENT. ' ') if LWP_USER_AGENT;
	my $response = $ua->get ($url);
	errorNoContents ("Error: communication failure - ". $response->status_line)
		unless $response->is_success;

	return $response->content;
}

;### Save the feed as a file
sub saveFeed {
	my ($hash, $feed) = @_;
	errorNoContents ("Program Exception: saveFeed [1]")
		unless defined $hash;
	errorNoContents ("Program Exception: saveFeed [2]")
		unless defined $feed;
;#
	use Fcntl ':flock';
	my $filename = getFilenameXML ($hash);
	if (open CACHE_WRITE, '+>'. $filename) {
		if (flock CACHE_WRITE, LOCK_EX) {
			seek CACHE_WRITE, 0,0;
			truncate CACHE_WRITE, 0;
			print CACHE_WRITE $$feed;
			flock CACHE_WRITE, LOCK_UN;
		}
		close FH;
	}
	else {
		errorNoContents ("Error: failed to open cache file - $hash");
	}
}

;### Solve the original URL from 2 hash numbers
sub retrieveURL {
	my ($next, $self) = @_;
	errorNoContents ("Program Exception: retrieveURL [1]")
		unless defined $next;
	errorNoContents ("Program Exception: retrieveURL [1]")
		unless defined $self;
;#
	my $fn_self = getFilenameXML ($self);
	return undef
		unless -e $fn_self;
	my $node_tree = getXMLDocumentTreeHash ($fn_self);

	;# Search URLs ... Damn!
	my @url = ();
	sub _recursiveSearch {
		my ($root_node) = @_;
		return unless ref ($root_node) eq 'HASH';
		foreach my $key (keys %$root_node) {
			push @url, $root_node->{$key}[0] if $key eq 'id';
			push @url, $root_node->{$key} if $key eq 'href';
			my $child_node = $root_node->{$key};
			next unless ref ($child_node) eq 'ARRAY';
			foreach my $child_tree (@$child_node) {
				_recursiveSearch ($child_tree);
			} 
		}
	}
	_recursiveSearch ($node_tree);

	;# Search specified URL in URLs ... Damn!!!
	foreach my $url (@url) {
		return $url if getURLHash ($url) eq $next;
	}
	return undef;
}

;### Retrieve startup URL
sub getStartupURL {
	my $filename = getFilenameXML ('0' x HASH_KEY_LENGTH);
	return undef unless -e $filename;
;#
	my $node_tree = getXMLDocumentTreeHash ($filename);
	return $node_tree && $node_tree->{feed}[0] ? $node_tree->{feed}[0]->{id}[0] : undef;
}

;### Process XML + XSLT
sub createOutput {
	my ($hash_xml, $hash_xsl) = @_;
;#
	;### Parse XML document with XML::LibXML
	my $xml_parser = XML::LibXML->new ()
		or errorNoContents ("Fatal Error: XML::LibXML initialize failure");
	my $doc_xml = eval { $xml_parser->parse_file (getFilenameXML ($hash_xml)) };
	errorNoContents ("Error: document parse failure at $hash_xml - $@")
		if $@;
	my $doc_xsl = eval { $xml_parser->parse_file (getFilenameXSL ($hash_xsl)) };
	errorNoContents ("Error: document parse failure at $hash_xsl - $@")
		if $@;

	;### Register user defined functions to XSLT engine
	my $xslt = XML::LibXSLT->new ()
		or errorNoContents ("Fatal Error: XML::LibXSLT initialize failure");
	$xslt->register_function (XMLNS_GGLCAL_FUNC, 'hashURL', \&getURLHash);
	$xslt->register_function (XMLNS_GGLCAL_FUNC, 'openURL', \&xslt_func_open_url);
	$xslt->register_function (XMLNS_GGLCAL_FUNC, 'reload', \&xslt_func_reload);
	$xslt->register_function (XMLNS_GGLCAL_FUNC, 'getPermalink', \&xslt_func_get_permalink);
	$xslt->register_function (XMLNS_GGLCAL_FUNC, 'goBack', \&xslt_func_go_back);

;#	$xslt->register_function ('urn:gglcalfunc', 'encodeHTML', \&encodeHTML);
;#	$xslt->register_function ('urn:gglcalfunc', 'decodeHTML', \&decodeHTML);
;#	$xslt->register_function ('urn:gglcalfunc', 'encodeXML', \&encodeXML);
;#	$xslt->register_function ('urn:gglcalfunc', 'decodeXML', \&decodeXML);

	;### Translation XML Document with XSLT
	my $xsl_obj = $xslt->parse_stylesheet ($doc_xsl)
		or errorNoContents ("Error: XSLT can't get stylesheet");
	my $results = $xsl_obj->transform ($doc_xml)
		or errorNoContents ("Error: XSLT can't transform XML document");

	return $xsl_obj->output_string ($results);
}

;########################################################################
;###	XSLT Procedures

;### Return a little JavaScripts to open the specified URL
sub xslt_func_open_url {
	my ($url, $mode) = @_;
;#
	return sprintf "javascript:void (doRequest ('%s','%s','%s'));",
			$mode ? $mode : $q_mode,
			$url ? getURLHash ($url) : '',
			$q_next;
}

;### Return a little JavaScripts to go to previous page.
sub xslt_func_go_back {
	my ($mode) = @_;
	return '' unless $q_self && $q_self ne $q_next;
;#
	return sprintf "javascript:void (doRequest ('%s','%s','%s'));",
			$mode ? $mode : $q_mode,
			$q_self, $q_next;
}

;### Return a little JavaScripts to reload myself
sub xslt_func_reload {
	return sprintf "javascript:void (doRequest (%s,%s,%s));",
			defined $q_mode ? "'$q_mode'" : "null",
			defined $q_next ? "'$q_next'" : "null",
			defined $q_self && $q_self ne $q_next ? "'$q_self'" : "null";
}

;### Return query strings for my permalink URL
sub xslt_func_get_permalink {
	my @param = ();
	push @param, "mode=$q_mode" if defined $q_mode;
	push @param, "next=$q_next" if defined $q_next;
	push @param, "self=$q_self" if defined $q_self && $q_self ne $q_next;
	return scalar @param ? '?'. join '&', @param : '';
}

;########################################################################
;###	Utilities

;### Encode the strings for URL (see decodeURL)
sub encodeURL {
	my ($str) = @_;
;#
	$str =~ s!([^a-zA-Z0-9_.-])!sprintf '%%%02X', ord ($1)!eg;
	return $str;
}

;### Decode the strings of URL (see encodeURL)
sub decodeURL {
	my ($str) = @_;
;#
	$str =~ s!%([0-9a-fA-F][0-9a-fA-F])!pack ('H*', $1)!eg;
	return $str;
}

;### Encode the strings of HTML (see decodeHTML)
sub encodeHTML {
	my ($html) = @_;
	return '' unless defined $html;
;#
	$html =~ tr!\cM!!d;
	;# Encode any & not followed by something that looks like
	;# an entity, numeric or otherwise.
	$html =~ s/&(?!#?[xX]?(?:[0-9a-fA-F]+|\w{1,8});)/&amp;/g;
	$html =~ s!"!&quot;!g;#"
	$html =~ s!<!&lt;!g;
	$html =~ s!>!&gt;!g;
	return $html;
}

;### Dencode the strings of HTML (see encodeHTML)
sub decodeHTML {
	my ($html) = @_;
	return '' unless defined $html;
;#
	$html =~ tr!\cM!!d;
	$html =~ s!&quot;!"!g;#"
	$html =~ s!&lt;!<!g;
	$html =~ s!&gt;!>!g;
	$html =~ s!&amp;!&!g;
	return $html;
}

;### Encode/Decode XML
{
	my %Map = ('&' => '&amp;', '"' => '&quot;', '<' => '&lt;', '>' => '&gt;', '\'' => '&apos;');
	my %Map_Decode = reverse %Map;
	my $RE = join '|', keys %Map;
	my $RE_D = join '|', keys %Map_Decode;

	;###
	sub encodeXML {
		my ($str, $nocdata) = @_;
		$nocdata ||= MT::ConfigMgr->instance->NoCDATA;
		if (!$nocdata && $str =~ m/
				<[^>]+>  ## HTML markup
				|        ## or
				&(?:(?!(\#([0-9]+)|\#x([0-9a-fA-F]+))).*?);
				## something that looks like an HTML entity.
				/x)
		{
			;# If ]]> exists in the string, encode the > to &gt;.
			$str =~ s/]]>/]]&gt;/g;
			$str = '<![CDATA[' . $str . ']]>';
		} else {
			$str =~ s!($RE)!$Map{$1}!g;
		}
		return $str;
	}

	;###
	sub decodeXML {
		my ($str) = @_;
		if ($str =~ s/<!\[CDATA\[(.*?)]]>/$1/g) {
			;# Decode encoded ]]&gt;
			$str =~ s/]]&(gt|#62);/]]>/g;
		} else {
			$str =~ s!($RE_D)!$Map_Decode{$1}!g;
		}
		return $str;
	}
}

;### Retrieve n bytes hex-strings of URL
sub getURLHash {
	my ($url) = @_;
;#
	;# md5_hex always returns lower-case
	return substr md5_hex (decodeURL (HASH_MAGICWORD. $url)), 0, HASH_KEY_LENGTH;
}

;### Return the hash referrence of XML document tree which specified by filename
sub getXMLDocumentTreeHash {
	my ($filename) = @_;
	errorNoContents ("Program Exception: getXMLDocumentTreeHash [1]")
		unless defined $filename;
;#
	my $xml_doc = XML::Simple->new ()
		or errorNoContents ("Fatal Error: XML::Simple initialize failure");
	my $node_tree = eval { $xml_doc->XMLin (
			$filename,
			KeepRoot => 1, ForceArray => 1) };
	errorNoContents ("Error: XML::Simple can't process $filename - $@")
		if $@;

	return $node_tree;
}

;########################################################################
__END__
